# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/7 14:45
# @Author : LucXiong
# @Project : Model
# @File : wps.py

"""
There are 2 kinds of wolf pack search algorithm.
The first one called wolf pack search(wps) proposed by Chenguang Yang, Xuyan Tu and Jie Chen in thier paper "Algorithm of Marriage in Honey Bees Optimization Based on the Wolf Pack Search" at 2007 International Conference on Intelligent Pervasive Computing in 2007. The wps code could ref: https://github.com/AlexanderKlanovets/swarm_algorithms/commit/78834820cadbcadb6902e3c171a2a8581255c542
The second one called wolf pack algorithm proposed by 吴虎胜,张凤鸣,吴庐山 in their paper "一种新的群体智能算法-狼群算法" at Systems Engineering and Electronics in 2013.
In fact, they are 2 different kinds of algorithm. WPS was shown in  this file, and WPA was in wpa.py.(I meet some problems in wpa.py, so it haven't been upload.)
"""

import numpy as np
import random
import matplotlib.pyplot as plt
import test_function

class wps():
    def __init__(self, n_dim=2, pop_size=50, max_iter=150, lb=-1e5, ub=1e5, step=0.5, func=None):
        # D - dimension of the search space;
        # N - number of wolves to generate;
        # step, l_near, t_max, beta - model parameters;
        # fitness_function - function to optimize;
        # w_range - range of particles' coordinates' values (from -range to range);
        # iter_num - maximum number of iterations.
        # (self, n_dim, pop_size, step, fitness_function, w_range, iter_num, l_near=0, t_max=0, beta=0):
        self.n_dim = n_dim
        self.pop = pop_size
        self.step = step # 影响朝最佳点靠近的速度
        self.func = func
        self.max_iter = max_iter

        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        assert self.n_dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # y = f(x) for all particles

        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_gbest()

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]

    def update(self):
        # 参考樽海鞘群算法
        # self.step = 2 * math.exp(-(4 * ((i + 1) / self.max_iter)) ** 2)
        for i in range(self.pop):
            if all(self.X[i] != self.gbest_x[0]): # self.Y[i] != self.gbest_y
                try:
                    self.X[i] += self.step * (self.gbest_x[0] - self.X[i]) / np.linalg.norm(self.gbest_x[0] - self.X[i])
                except:
                    self.X[i] += self.step * (self.gbest_x - self.X[i]) / np.linalg.norm(self.gbest_x - self.X[i])
        self.X = np.clip(self.X, self.lb, self.ub)
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # Function for fitness evaluation of new solutions

    def run(self):
        for iter in range(self.max_iter):
            self.update() # self.step可以选择参考樽海鞘群，随迭代次数发生改变
            self.update_pbest()
            self.update_gbest()
            self.gbest_y_hist.append(self.gbest_y)
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y

if __name__ == '__main__':
    n_dim = 2
    lb = [-5.12 for i in range(n_dim)]
    ub = [5.12 for i in range(n_dim)]
    demo_func = test_function.f22
    ssa = wps(pop_size=50, n_dim=n_dim, max_iter=300, lb=lb, ub=ub, func=demo_func, step=0.5)
    ssa.run()
    print('best_x is ', ssa.gbest_x, 'best_y is', ssa.gbest_y)
    print(f'{demo_func(ssa.gbest_x)}\t{ssa.gbest_x}')
    plt.plot(ssa.gbest_y_hist)
    plt.show()


# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/15 19:57
# @Author : LucXiong
# @Project : Model
# @File : __init__.py

from .CSA import CSA
from .FA import FA
from .PSO import PSO
from .SBO import SBO
from .SCA import sca as SCA
from .SSA2017 import salp_swarm_algorithm as SSA2017
from .SSA2020 import SSA as SSA2020
from .WPA import wpa as WPA
from .WPS import wps as WPS



# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/15 15:02
# @Author : LucXiong
# @Project : Model
# @File : FA.py

"""
Ref:https://github.com/GoodLittleStar/Fireworks/blob/master/FireWork.py
Ref:Tan Y, Zhu Y. Fireworks Algorithm for Optimization[M].
Lecture Notes in Computer Science. City: Springer Berlin Heidelberg, 2010: 355-64[2021-12-08T08:42:21].
"""

import random  # random Function
import numpy as np # numpy operations
import copy
import matplotlib.pyplot as plt
import math
import test_function

class FA():
    def __init__(self, pop_size=50, n_dim=2, m=50, a=0.04, b=0.8, A=40, lb=-1e5, ub=1e5, max_iter=1000, func=None):

        self.a = a
        self.b = b
        self.m = m
        self.A = A
        self.pop = pop_size
        self.dim = n_dim
        self.func = func
        self.max_iter = max_iter  # max iter
        self.epsino = 1e-6
        self.mutate = 0.1

        self.lb, self.ub = np.array(lb) * np.ones(self.dim), np.array(ub) * np.ones(self.dim)
        assert self.dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.dim))
        self.X = self.X.tolist()
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # y = f(x) for all particles

        # self.pbest_x = self.X.copy()  # personal best location of every particle in history
        # self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        # self.fit = [1 / (1 + self.Y[i]) if self.Y[i] > 0 else 1 - self.Y[i] for i in range(self.pop)]
        # self.prob = [self.fit[i] / sum(self.fit) for i in range(self.pop)]
        self.bestindex = self.Y.index(min(self.Y))
        self.gbest_x = self.X[self.bestindex]
        self.gbest_y = min(self.Y) # global best y for all particles
        self.gbest_y_hist = [self.gbest_y]  # gbest_y of every iteration
        # self.update_gbest()


    # def update_pbest(self):
    #     '''
    #     personal best
    #     :return:
    #     '''
    #     for i in range(len(self.Y)):
    #         if self.pbest_y[i] > self.Y[i]:
    #             self.pbest_x[i] = self.X[i]
    #             self.pbest_y[i] = self.Y[i]
    #
    # def update_gbest(self):
    #     '''
    #     global best
    #     :return:
    #     '''
    #     idx_min = self.pbest_y.index(min(self.pbest_y))
    #     if self.gbest_y > self.pbest_y[idx_min]:
    #         self.gbest_x = self.X[idx_min, :].copy()
    #         self.gbest_y = self.pbest_y[idx_min]

    def CalculateSi(self):
        self.MaxFitness = max(self.Y)
        temp = 0.
        self.Si = []
        for i in range(0, self.pop):
            temp = temp + self.MaxFitness - self.Y[i]
        for i in range(0, self.pop):
            self.Si.append(self.m * (self.MaxFitness - self.Y[i] + self.epsino) / (temp + self.epsino))
            if self.Si[-1] < self.a * self.m:
                self.Si[-1] = round(self.a * self.m)
            elif self.Si[-1] > self.b * self.m:
                self.Si[-1] = round(self.b * self.m)
            else:
                self.Si[-1] = round(self.Si[-1])

    def CalculateExpo(self):
        self.MinFitness = min(self.Y)
        temp = 0.
        self.Ai = []
        for i in range(self.pop):
            temp = temp + self.Y[i] - self.MinFitness
        for i in range(self.pop):
            self.Ai.append(self.A * (self.Y[i]- self.MinFitness + self.epsino) / (temp + self.epsino))

    def Explosion(self):
        for k in range(0, self.pop):
            for i in range(self.Si[k]):
                spark = copy.deepcopy(self.X[k])
                z = round(self.dim * random.uniform(0, 1))
                dim_list = range(self.dim)
                rand_z = random.sample(dim_list, z)
                h = self.Ai[k] * random.uniform(-1, 1)
                for j in rand_z:
                    spark[j] += h
                    if spark[j] < self.lb[j] or spark[j] > self.ub[j]:
                        spark[j] = self.lb[j] + abs(spark[j]) % (self.ub[j] - self.lb[j])
                self.X.append(spark)
                self.Y.append(self.func(spark))
            if(len(self.X) > 5 * self.pop):
                break

    def Mutation(self):
        currentsize = len(self.X)
        for k in range(round(self.mutate * currentsize)):
            randindex = random.randint(0, currentsize - 1)
            spark = copy.deepcopy(self.X[randindex])
            # print(spark)
            # print(randindex)
            z = round(self.dim * random.uniform(0, 1))
            dim_list = range(self.dim)
            rand_z = random.sample(dim_list, z)
            g = random.gauss(1, 1)
            for j in rand_z:
                spark[j] *= g
                if spark[j] < self.lb[j] or spark[j] > self.ub[j]:
                    spark[j] = self.lb[j] + abs(spark[j]) % (self.ub[j] - self.lb[j])
            self.X.append(spark)
            self.Y.append(self.func(spark))
            if (len(self.X) > 10 * self.pop):
                break

    def Selection(self):
        newpop=[]
        newpop.append(self.gbest_x)
        self.Ri = []
        for i in range(len(self.X)):
            dis=0.
            for j in range(len(self.X)):
                for k in range(self.dim):
                    dis+= (self.X[i][k]-self.X[j][k])**2
            self.Ri.append(math.sqrt(dis))
        sr = sum(self.Ri)
        px = [self.Ri[i]/sr for i in range(len(self.Ri))]
        for i in range(self.pop-1):
            rr=random.uniform(0,1)
            index=0
            for j in range(self.pop):
                if j==0 and rr<px[j]:
                    index=j
                elif rr>=px[j] and rr<px[j+1]:
                    index=j+1
            newpop.append(self.X[index])
        self.X = newpop
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]

    def run(self):
        for i in range(self.max_iter):
            print(i)
            print(len(self.X))
            self.CalculateSi()
            self.CalculateExpo()
            self.Explosion()
            print(len(self.X))
            self.Mutation()
            print(len(self.X))
            bestindex = self.Y.index(min(self.Y))
            if self.gbest_y_hist[-1] > self.Y[bestindex]:
                self.gbest_y_hist.append(self.Y[bestindex])
                self.gbest_x = self.X[bestindex]
            else:
                self.gbest_y_hist.append(self.gbest_y_hist[-1])
            self.Selection()
            print(self.gbest_y_hist[-1])
        return self.gbest_x, self.gbest_y_hist[-1]

if __name__ == '__main__':
    n_dim = 30
    lb = [-100 for i in range(n_dim)]
    ub = [100 for i in range(n_dim)]
    demo_func = test_function.fm2
    pop_size = 100
    max_iter = 200
    fa = FA(n_dim=n_dim, pop_size=pop_size, max_iter=max_iter, lb=lb, ub=ub, func=demo_func)
    best_x, bext_y = fa.run()
    print(f'{demo_func(fa.gbest_x)}\t{fa.gbest_x}')
    plt.plot(fa.gbest_y_hist)
    plt.show()



# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/15 15:57
# @Author : LucXiong
# @Project : Model
# @File : PSO.py

"""
Ref:https://github.com/guofei9987/scikit-opt
已经有了很多可以调用粒子群算法的库，但是在高维情形下的寻优效果很差，所以决定自己再写一个
"""

import numpy as np
import matplotlib.pyplot as plt
import math
import test_function

class PSO():
    def __init__(self, func, n_dim=None, pop=40, max_iter=150, lb=-1e5, ub=1e5, w=0.8, c1=0.5, c2=0.5):
        self.func = func
        self.w = w  # inertia
        self.cp, self.cg = c1, c2  # parameters to control personal best, global best respectively
        self.pop = pop  # number of particles
        self.n_dim = n_dim  # dimension of particles, which is the number of variables of func
        self.max_iter = max_iter  # max iter

        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        assert self.n_dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))
        v_high = self.ub - self.lb
        self.V = np.random.uniform(low=-v_high, high=v_high, size=(self.pop, self.n_dim))  # speed of particles
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # y = f(x) for all particles

        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_gbest()

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]

    def update_V(self):
        r1 = np.random.rand(self.pop, self.n_dim)
        r2 = np.random.rand(self.pop, self.n_dim)
        self.V = self.w * self.V + self.cp * r1 * (self.pbest_x - self.X) + self.cg * r2 * (self.gbest_x - self.X)

    def update(self):
        for i in range(self.pop):
            self.X[i] += self.V[i]
        self.X = np.clip(self.X, self.lb, self.ub)
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]

    def run(self):
        for iter_no in range(self.max_iter):
            self.update_V()
            self.update()
            self.update_pbest()
            self.update_gbest()
            self.gbest_y_hist.append(self.gbest_y)
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y

if __name__ == '__main__':
    n_dim = 30
    lb = [-100 for i in range(n_dim)]
    ub = [100 for i in range(n_dim)]
    demo_func = test_function.fu2
    pop_size = 100
    max_iter = 1000
    pso = PSO(func=demo_func, n_dim=n_dim, pop=100, max_iter=1000, lb=lb, ub=ub, w=0.8, c1=0.5, c2=0.5)
    best_x, bext_y = pso.run()
    print(f'{demo_func(pso.gbest_x)}\t{pso.gbest_x}')
    # plt.plot(pso.gbest_y_hist)
    # plt.show()


# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/7 14:06
# @Author : LucXiong
# @Project : Model
# @File : wpa.py

"""
There are 2 kinds of wolf pack search algorithm.
The first one called wolf pack search(wps) proposed by Chenguang Yang, Xuyan Tu and Jie Chen in thier paper "Algorithm of Marriage in Honey Bees Optimization Based on the Wolf Pack Search" at 2007 International Conference on Intelligent Pervasive Computing in 2007. The wps code could ref: https://github.com/AlexanderKlanovets/swarm_algorithms/commit/78834820cadbcadb6902e3c171a2a8581255c542
The second one called wolf pack algorithm proposed by 吴虎胜,张凤鸣,吴庐山 in their paper "一种新的群体智能算法-狼群算法" at Systems Engineering and Electronics in 2013.
In fact, they are 2 different kinds of algorithm. WPA was shown in  this file, and WPS was in wps.py.

Ref:git@github.com:xzltc/WPA.git
Ref:https://blog.csdn.net/weixin_37978667/article/details/85112709
>吴虎胜,张凤鸣,吴庐山.一种新的群体智能算法——狼群算法[J].系统工程与电子技术, 2013, 35(11): 2430-2438.
>周强,周永权-种基于领导者策略的狼群搜索算法[J.计算机应用研究,2013,30(9):2629-2632.
"""
import numpy as np
import math
import random
import heapq
from test_func import *
import matplotlib.pyplot as plt


class wpa():
    def __init__(self, pop_size=50, n_dim=2, alpha=4, beta=6, w=100, lb=-1e5, ub=1e5, max_iter=300, func=None):
        self.pop = pop_size
        self.n_dim = n_dim
        self.func = func
        self.max_iter = max_iter  # max iter
        self.alpha = alpha  # 探狼比例因子 （取[n／(α+1)，n／α]之间的整数）
        self.beta = beta  # 狼群更新比例因子
        self.w = w  # 距离判定因子
        self.S = 1000  # 步长因子
        self.number_T = np.random.randint(self.pop / (self.alpha + 1), self.pop / self.alpha)  # 探狼数量
        self.h = 10 # 探狼探寻方向总数
        self.T_max = 30  # 最大游走次数

        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        assert self.n_dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.step_a = (self.ub - self.lb) / self.S  # 探狼游走步长
        self.step_b = 2 * self.step_a  # 猛狼奔袭步长
        self.step_c = self.step_a / 2  # 围攻步长

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # y = f(x) for all particles

        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [self.Y[i] for i in range(self.pop)]  # best image of every particle in history
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_gbest()

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        # self.need_update = self.pbest_y > self.Y
        # print(self.need_update)
        # print(self.pbest_y)
        # print(self.Y)
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]
        # self.pbest_x = np.where(self.need_update, self.X, self.pbest_x)
        # self.pbest_y = np.where(self.need_update, self.Y, self.pbest_y)
        # print(self.pbest_x)

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]

    def update(self):
        L_Wolf_Y = np.min(self.Y)  # 记录头狼的值
        L_Wolf_index = np.argmin(self.Y)  # 头狼在狼群(wol)f_colony_X)中下标位置
        L_Wolf_X = self.X[L_Wolf_index]  # 记录头狼的变量信息
        L_Wolf = [L_Wolf_Y, L_Wolf_index]

        # 初始化探狼
        T_Wolf_V = heapq.nsmallest(self.number_T, self.Y.copy())
        T_Wolf = self.find_index(T_Wolf_V, self.Y)  # 0:探狼在狼群中位置 1:解
        # 初始化猛狼
        M_Wolf_V = heapq.nlargest(self.pop - self.number_T - 1, self.Y.copy())
        M_Wolf = self.find_index(M_Wolf_V, self.Y)  # 0:探狼在狼群中位置 1:解
        self.detective_wolf(T_Wolf, L_Wolf_Y, L_Wolf_X, L_Wolf)
        # 需要修改，参考一种新的群体智能算法 狼群算法> 2013.11 式（3）
        # 言w增大会加速算法收敛，但W过大会使得人工狼很难进入围攻行为，缺乏对猎物的精细搜索。
        d_near = 0
        for d in range(self.n_dim):
            d_near += ((1 / (1 * self.w * self.n_dim)) * (self.ub[d] - self.lb[d]))  # dnear值，判定距离
        self.call_followers(d_near, M_Wolf, L_Wolf_Y, L_Wolf_X, L_Wolf_index)
        self.siege(M_Wolf, L_Wolf_X, L_Wolf_Y, L_Wolf_index)
        self.live()

    def find_index(self, value, wolf_colony):
        ret = np.zeros((len(value), 2))

        for i in range(0, len(value)):
            ret[i][0] = wolf_colony.index(value[i])
            ret[i][1] = value[i]
        return ret

    def detective_wolf(self, T_Wolf, L_Wolf_Y, L_Wolf_X, L_Wolf):
        # 探狼开始游走
        for i in range(0, T_Wolf.shape[0]):
            H = np.random.randint(2, self.h)  # 尝试的方向
            single_T_Wolf = self.X[int(T_Wolf[i][0])]  # 当前探狼
            optimum_value = T_Wolf[i][1]  # 游走的最优解
            optimum_position = single_T_Wolf.copy()  # 游走最优位置
            find = False

            # 探狼游走行为 一旦探狼发现目标 -> 开始召唤
            for t in range(0, self.T_max):
                # 在初始位置朝H个方向试探，直到找到一个优解
                for p in range(1, H + 1):
                    single_T_Wolf_trial = single_T_Wolf.copy()
                    # 一种新的群体智能算法 狼群算法> 2013.11 式（1）
                    # h越大探狼搜寻得越精细但同时速度也相对较慢
                    # 这个部分有点像果蝇的搜素
                    single_T_Wolf_trial = single_T_Wolf_trial + math.sin(2 * math.pi * p / H) * self.step_a # 根据原式不需要转化类型为int
                    # 一种基于领导者策略的狼群搜索算法> 2013.9 式（2）
                    # single_T_Wolf_trial = single_T_Wolf_trial + np.random.uniform(-1, 1,(single_T_Wolf_trial.shape[0],)) * self.step_a

                    single_T_Wolf_V = self.func(single_T_Wolf_trial) # [0]

                    # 探狼转变为头狼
                    if L_Wolf_Y > single_T_Wolf_V:
                        find = True
                        L_Wolf_Y = single_T_Wolf_V  # 更新头狼解
                        L_Wolf_X = single_T_Wolf_trial
                        L_Wolf_index = int(T_Wolf[i][0])  # 更新头狼下标
                        self.X[L_Wolf_index] = single_T_Wolf_trial  # 更新头狼位置参数

                        T_Wolf = np.delete(T_Wolf, i, axis=0)  # 探狼转为头狼，发起召唤，删除探狼
                        break

                    elif optimum_value > single_T_Wolf_V:
                        optimum_value = single_T_Wolf_V
                        optimum_position = single_T_Wolf_trial

                else:
                    # print(" > 第%d只探狼完成第%d次游走,未发现猎物" % ((i + 1), (t + 1))) # 记录上次的最优位置
                    single_T_Wolf = optimum_position

                if find is True:
                    break

            if find is True:
                print("第%d只探狼发现猎物 %f" % (i, single_T_Wolf_V))
                break

            else:
                # 若游走完成探狼没找到猎物，更新所有游走过程中最优的一次位置
                self.X[int(T_Wolf[i][0])] = optimum_position

    def call_followers(self, d_near, M_Wolf, L_Wolf_Y, L_Wolf_X, L_Wolf_index):
        # 召唤行为
        surrounded = False
        # 所有猛狼进入围攻范围才能结束
        while ~surrounded:
            ready = 0
            for m in range(0, M_Wolf.shape[0]):
                s_m_index = int(M_Wolf[m][0])  # 猛狼在狼群中下标
                single_M_Wolf = self.X[s_m_index]  # 猛狼变量
                d = np.abs(L_Wolf_X - single_M_Wolf)
                dd = np.sum(d)

                while d_near < dd:
                    for d in range(self.n_dim):
                        single_M_Wolf[d] = single_M_Wolf[d] + self.step_b[d] * (L_Wolf_X[d] - single_M_Wolf[d]) / np.abs(L_Wolf_X[d] - single_M_Wolf[d])
                    single_M_Wolf_V = self.func(single_M_Wolf)

                    # 更新猛狼位置
                    self.X[s_m_index] = single_M_Wolf
                    M_Wolf[m][1] = single_M_Wolf_V

                    d = np.abs(L_Wolf_X - single_M_Wolf)
                    dd = np.sum(d)

                    if L_Wolf_Y > single_M_Wolf_V:
                        # 头狼变猛狼
                        M_Wolf[m][0] = L_Wolf_index
                        M_Wolf[m][1] = L_Wolf_Y
                        self.X[L_Wolf_index] = L_Wolf_X

                        # 猛狼变头狼
                        L_Wolf_Y = single_M_Wolf_V
                        L_Wolf_X = single_M_Wolf
                        L_Wolf_index = s_m_index
                        self.X[L_Wolf_index] = single_M_Wolf
                        break

                if d_near > dd:
                    ready += 1  # 围攻就绪态+1
                else:
                    break

            # 所有猛狼是否就位
            if ready == M_Wolf.shape[0]:
                print(" > 所有猛狼已进入围攻状态")
                break

    def siege(self, M_Wolf, L_Wolf_X, L_Wolf_Y, L_Wolf_index):
        for m in range(0, M_Wolf.shape[0]):
            s_m_index = int(M_Wolf[m][0])
            single_M_Wolf = self.X[s_m_index]
            # 发起围攻，计算围攻后位置
            single_M_Wolf = single_M_Wolf + np.random.uniform(-1, 1) * self.step_c * np.abs(L_Wolf_X - single_M_Wolf)
            single_M_Wolf_V = self.func(single_M_Wolf)

            if L_Wolf_Y > single_M_Wolf_V:
                print(" > 发起围攻!目标更新   原值:%f => 现值:%f" % (L_Wolf_Y, single_M_Wolf_V))
                # 头狼变猛狼
                M_Wolf[m][0] = L_Wolf_index
                M_Wolf[m][1] = L_Wolf_Y
                self.X[L_Wolf_index] = L_Wolf_X

                # 猛狼变头狼
                L_Wolf_Y = single_M_Wolf_V
                L_Wolf_X = single_M_Wolf
                L_Wolf_index = s_m_index
                self.X[L_Wolf_index] = single_M_Wolf
        if self.gbest_y > single_M_Wolf_V:
            self.gbest_y = single_M_Wolf_V
            self.gbest_x = single_M_Wolf
            self.gbest_y_hist.append(self.gbest_y)
        else:
            self.gbest_y_hist.append(self.gbest_y)


    def live(self):
        # 强者生存行为
        wolf_colony_V = [self.func(self.X[i]) for i in range(len(self.X))]  # 重新计算现在所有狼狩猎的状态
        eliminate_number = np.random.randint(self.pop / (2 * self.beta), self.pop / self.beta)
        Bad_Wolf_V = heapq.nlargest(eliminate_number, wolf_colony_V.copy())
        Bad_Wolf = self.find_index(Bad_Wolf_V, wolf_colony_V)
        # 生成新的人工狼
        new_wolf = np.random.uniform(self.lb, self.ub, (eliminate_number, self.n_dim))
        # 淘汰最弱的狼
        for n in range(0, len(Bad_Wolf_V)):
            self.X[int(Bad_Wolf[n][0])] = new_wolf[n]

    def run(self):
        for i in range(self.max_iter):
            print(i)
            self.update()
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y

if __name__ == '__main__':
    n_dim = 2
    lb = [-5.12 for i in range(n_dim)]
    ub = [5.12 for i in range(n_dim)]
    demo_func = f22
    #  def __init__(self, pop_size=50, n_dim=2, alpha=4, beta=6, w=500, lb=-1e5, ub=1e5, max_iter=300, func=None):
    wpa = wpa(pop_size=50, n_dim=n_dim, max_iter=1000, lb=lb, ub=ub, func=demo_func, alpha=4, beta=6, w=500)
    wpa.run()
    print('best_x is ', wpa.gbest_x, 'best_y is', wpa.gbest_y)
    # f22 is best_x is  [ 0.09110638 -0.71978079] best_y is -1.0312114438653428
    # f23 is best_x is  [-448.40744098  403.16544094] best_y is -894.7260832993893 效果不行
    print(f'{demo_func(wpa.gbest_x)}\t{wpa.gbest_x}')
    plt.plot(wpa.gbest_y_hist)
    plt.show()



# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/5 15:39
# @Author : LucXiong
# @Project : Model
# @File : SSA2017.py

############################################################################

# Ref: Salp Swarm Algorithm: A bio-inspired optimizer for engineering design problems.pdf

############################################################################

# Required Libraries
import numpy  as np
import math
import random
import os
import matplotlib.pyplot as plt
import test_function

class salp_swarm_algorithm():
    def __init__(self, pop_size=50, n_dim=2, max_iter=150, lb=[-5,-5], ub=[5,5], func=None):
        self.pop = pop_size
        self.lb = lb
        self.ub = ub
        self.func = func
        self.n_dim = n_dim
        self.max_iter = max_iter


        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # y = f(x) for all particles
        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_pbest()
        self.update_gbest()

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]



    # Function: Updtade Position
    def update_position(self, c1):
        for i in range(0, self.pop):
            if (i <= self.pop / 2): # 领导者比例
                for j in range(0, self.n_dim):
                    c2 = int.from_bytes(os.urandom(8), byteorder="big") / ((1 << 64) - 1)
                    c3 = int.from_bytes(os.urandom(8), byteorder="big") / ((1 << 64) - 1)
                    if (c3 >= 0.5):  # c3 < 0.5
                        try:
                            self.X[i, j] = np.clip((self.gbest_x[0][j] + c1 * ((self.ub[j] - self.lb[j]) * c2 + self.lb[j])), self.lb[j], self.ub[j])
                        except:
                            self.X[i, j] = np.clip((self.gbest_x[j] + c1 * ((self.ub[j] - self.lb[j]) * c2 + self.lb[j])), self.lb[j], self.ub[j])
                    else:
                        try:
                            self.X[i, j] = np.clip((self.gbest_x[0][j] - c1 * ((self.ub[j] - self.lb[j]) * c2 + self.lb[j])), self.lb[j], self.ub[j])
                        except:
                            self.X[i, j] = np.clip((self.gbest_x[j] - c1 * ((self.ub[j] - self.lb[j]) * c2 + self.lb[j])), self.lb[j], self.ub[j])
            else: # 追随者比例
                for j in range(0, self.n_dim):
                    self.X[i, j] = np.clip(((self.X[i - 1, j] + self.X[i, j]) / 2), self.lb[j], self.ub[j])
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # y = f(x) for all particles

    def run(self):
        for i in range(self.max_iter):
            c1 = 2 * math.exp(-(4 * ((i+1) / self.max_iter)) ** 2)
            self.update_position(c1)
            self.update_pbest()
            self.update_gbest()
            self.gbest_y_hist.append(self.gbest_y)
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y


if __name__ == '__main__':
    n_dim = 30
    lb = [-5 for i in range(n_dim)]
    ub = [5 for i in range(n_dim)]
    demo_func = test_function.fu1
    ssa = salp_swarm_algorithm(pop_size=50, n_dim=n_dim, max_iter=150, lb=lb, ub=ub, func=demo_func)
    ssa.run()
    print('best_x is ', ssa.gbest_x, 'best_y is', ssa.gbest_y)
    print(f'{demo_func(ssa.gbest_x)}\t{ssa.gbest_x}')
    plt.plot(ssa.gbest_y_hist)
    plt.show()






# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/2 11:10
# @Author : LucXiong
# @Project : Model
# @File : SSA.py

"""
Ref:https://github.com/changliang5811/SSA_python
Ref:https://www.tandfonline.com/doi/full/10.1080/21642583.2019.1708830
Ref:A novel swarm intelligence optimization approach: sparrow search algorithm.pdf
"""

import numpy as np
import matplotlib.pyplot as plt
import math
import test_function

class SSA():
    def __init__(self, func, n_dim=None, pop_size=20, max_iter=50, lb=-512, ub=512, verbose=False):
        self.func = func
        self.n_dim = n_dim  # dimension of particles, which is the number of variables of func
        self.pop = pop_size  # number of particles
        P_percent = 0.2 # # 生产者的人口规模占总人口规模的20%
        D_percent = 0.1 # 预警者的人口规模占总人口规模的10%
        self.pNum = round(self.pop * P_percent)  # 生产者的人口规模占总人口规模的20%
        self.warn = round(self.pop * D_percent)  # 预警者的人口规模占总人口规模的10%

        self.max_iter = max_iter  # max iter
        self.verbose = verbose  # print the result of each iter or not

        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        assert self.n_dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))

        self.Y = [self.func(self.X[i]) for i in range(len(self.X))] # y = f(x) for all particles
        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_pbest()
        self.update_gbest()
        #
        # record verbose values
        self.record_mode = False
        self.record_value = {'X': [], 'V': [], 'Y': []}
        self.best_x, self.best_y = self.gbest_x, self.gbest_y  # history reasons, will be deprecated
        self.idx_max = 0
        self.x_max = self.X[self.idx_max, :]
        self.y_max = self.Y[self.idx_max]

    def cal_y(self, start, end):
        for i in range(start, end):
            self.Y[i] = self.func(self.X[i])

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]

    def find_worst(self):
        self.idx_max = self.Y.index(max(self.Y))
        self.x_max = self.X[self.idx_max, :]
        self.y_max = self.Y[self.idx_max]

    def update_finder(self):
        r2 = np.random.rand(1)  # 预警值
        self.idx = sorted(enumerate(self.Y), key=lambda x: x[1])
        self.idx = [self.idx[i][0] for i in range(len(self.idx))]
        # 这一部位为发现者（探索者）的位置更新
        if r2 < 0.8:  # 预警值较小，说明没有捕食者出现
            for i in range(self.pNum):
                r1 = np.random.rand(1)
                self.X[self.idx[i], :] = self.X[self.idx[i], :] * np.exp(-(i) / (r1 * self.max_iter))  # 对自变量做一个随机变换
                self.X = np.clip(self.X, self.lb, self.ub) # 对超过边界的变量进行去除
                # X[idx[i], :] = Bounds(X[idx[i], :], lb, ub)  # 对超过边界的变量进行去除
                # fit[sortIndex[0, i], 0] = func(X[sortIndex[0, i], :])  # 算新的适应度值
        elif r2 >= 0.8:  # 预警值较大，说明有捕食者出现威胁到了种群的安全，需要去其它地方觅食
            for i in range(self.pNum):
                Q = np.random.rand(1)  # 也可以替换成  np.random.normal(loc=0, scale=1.0, size=1)
                self.X[self.idx[i], :] = self.X[self.idx[i], :] + Q * np.ones((1, self.n_dim))  # Q是服从正态分布的随机数。L表示一个1×d的矩阵
                self.X = np.clip(self.X, self.lb, self.ub)  # 对超过边界的变量进行去除
                # X[idx[i], :] = Bounds(X[sortIndex[0, i], :], lb, ub)
                # fit[sortIndex[0, i], 0] = func(X[sortIndex[0, i], :])
        self.cal_y(0, self.pNum)

    def update_follower(self):
        #  这一部位为加入者（追随者）的位置更新
        for ii in range(self.pop - self.pNum):
            i = ii + self.pNum
            A = np.floor(np.random.rand(1, self.n_dim) * 2) * 2 - 1
            best_idx = self.Y[0:self.pNum].index(min(self.Y[0:self.pNum]))
            bestXX = self.X[best_idx, :]
            if i > self.pop/2: 
                Q = np.random.rand(1)
                self.X[self.idx[i],:] = Q*np.exp((self.x_max-self.X[self.idx[i],:])/np.square(i))
            else: 
                self.X[self.idx[i],:] = bestXX+np.dot(np.abs(self.X[self.idx[i],:]-bestXX),1/(A.T*np.dot(A,A.T)))*np.ones((1, self.n_dim))
        self.X = np.clip(self.X, self.lb, self.ub)  # 对超过边界的变量进行去除
        self.cal_y(self.pNum, self.pop)

    def detect(self):
        arrc = np.arange(self.pop)
        c = np.random.permutation(arrc)  # 随机排列序列
        b = [self.idx[i] for i in c[0: self.warn]]
        e = 10e-10
        for j in range(len(b)):
            if self.Y[b[j]] > self.gbest_y:
                self.X[b[j], :] = self.gbest_y + np.random.rand(1, self.n_dim) * np.abs(self.X[b[j], :] - self.gbest_y)
            else:
                self.X[b[j], :] = self.X[b[j], :]+(2*np.random.rand(1)-1)*np.abs(self.X[b[j], :]-self.x_max)/(self.func(self.X[b[j]])-self.y_max+e)
            self.X = np.clip(self.X, self.lb, self.ub)
            self.Y[b[j]] = self.func(self.X[b[j]])

    def run(self, max_iter=None):
        self.max_iter = max_iter or self.max_iter
        for iter_num in range(self.max_iter):
            self.update_finder() # 更新发现者位置
            self.find_worst()  # 取出最大的适应度值和最差适应度的X
            self.update_follower() # 更新跟随着位置
            self.update_pbest()
            self.update_gbest()
            self.detect()
            self.update_pbest()
            self.update_gbest()
            self.gbest_y_hist.append(self.gbest_y)
        return self.best_x, self.best_y

if __name__ == '__main__':
    n_dim = 30
    lb = [-100 for i in range(n_dim)]
    ub = [100 for i in range(n_dim)]
    demo_func = test_function.fu5
    pop_size = 100
    max_iter = 100
    ssa = SSA(demo_func, n_dim=n_dim, pop_size=pop_size, max_iter=max_iter, lb=lb, ub=ub)
    ssa.run()
    print('best_x is ', ssa.gbest_x, 'best_y is', ssa.gbest_y)
    # print(f'{demo_func(ssa.gbest_x)}\t{ssa.gbest_x}')
    plt.plot(ssa.gbest_y_hist)
    plt.show()



# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/3 15:23
# @Author : LucXiong
# @Project : Model
# @File : CSA.py

"""
Ref:
[1] https://github.com/Luan-Michel/CrowSearchAlgorithmPython/blob/master
# /CrowSearchAlgorithm.py
[2] Alireza Askarzadeh, Anovel metaheuristic method for solving constrained
# engineering optimization problems: Crow search algorithm, Computers &
# Structures, Vol. 169, 1-12, 2016.
"""

import random  # random Function
import numpy as np # numpy operations
import matplotlib.pyplot as plt
import math  # ceil function
import test_function

class CSA():
    def __init__(self, pop_size=5, n_dim=2, ap=0.1, lb=-1e5, ub=1e5, max_iter=20, func=None):
        self.pop = pop_size
        self.n_dim = n_dim
        self.ap = ap # 感知概率
        self.func = func
        self.max_iter = max_iter  # max iter
        self.fly_length = [2 for _ in range(self.n_dim)] # 飞行距离，可以考虑是否采用莱维飞行或者随迭代次数改变
        # 或许也和变量维数相关
        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        assert self.n_dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))] # y = f(x) for all particles

        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_gbest()

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]

    def update(self):
        num = np.array([random.randint(0, self.pop - 1) for _ in range(self.pop)])  # Generation of random candidate crows for following (chasing)
        for i in range(self.pop):
            if (random.random() > self.ap):
                for j in range(self.n_dim):
                    self.X[(i, j)] = self.X[(i, j)] + self.fly_length[j] * ((random.random()) * (self.pbest_x[(num[i], j)] - self.X[(i, j)]))
            else:
                for j in range(self.n_dim): # 随机生成或许可以考虑采用莱维飞行
                    self.X[(i, j)] = self.lb[j] - (self.lb[j] - self.ub[j]) * random.random()
        self.X = np.clip(self.X, self.lb, self.ub)
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # Function for fitness evaluation of new solutions

    def run(self):
        for iter in range(self.max_iter):
            self.update()
            self.update_pbest()
            self.update_gbest()
            self.gbest_y_hist.append(self.gbest_y)
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y

if __name__ == '__main__':
    # todo(xionglei@sjtu.edu.cn): 哪里有问题，复现的寻优效果很差
    n_dim = 30
    lb = [-100 for i in range(n_dim)]
    ub = [100 for i in range(n_dim)]
    demo_func = test_function.fu2
    pop_size = 100
    max_iter = 1000
    csa = CSA(n_dim=n_dim, pop_size=pop_size, max_iter=max_iter, lb=lb, ub=ub, func=demo_func)
    best_x, bext_y = csa.run()
    print(f'{demo_func(csa.gbest_x)}\t{csa.gbest_x}')
    plt.plot(csa.gbest_y_hist)
    plt.show()




# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/8 14:20
# @Author : LucXiong
# @Project : Model
# @File : SCA_new.py

"""
Ref:https://github.com/luizaes/sca-algorithm
S. Mirjalili, SCA: A Sine Cosine Algorithm for Solving Optimization Problems, Knowledge-based Systems, in press, 2015, DOI: http://dx.doi.org/10.1016/j.knosys.2015.12.022
"""

import numpy as np
import random
import math
import matplotlib.pyplot as plt
import test_function

class sca():
    def __init__(self, pop_size=5, n_dim=2, a=2, lb=-1e5, ub=1e5, max_iter=20, func=None):
        self.pop = pop_size
        self.n_dim = n_dim
        self.a = a # 感知概率
        self.func = func
        self.max_iter = max_iter  # max iter

        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        assert self.n_dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))] # y = f(x) for all particles

        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_gbest()

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]


    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]

    def update(self, i):
        r1 = self.a - i * ((self.a) / self.max_iter)
        for j in range(self.pop):
            for k in range(self.n_dim):
                r2 = 2 * math.pi * random.uniform(0.0, 1.0)
                r3 = 2 * random.uniform(0.0, 1.0)
                r4 = random.uniform(0.0, 1.0)
                if r4 < 0.5:
                    try:
                        self.X[j][k] = self.X[j][k] + (r1 * math.sin(r2) * abs(r3 * self.gbest_x[k] - self.X[j][k]))
                    except:
                        self.X[j][k] = self.X[j][k] + (r1 * math.sin(r2) * abs(r3 * self.gbest_x[0][k] - self.X[j][k]))
                else:
                    try:
                        self.X[j][k] = self.X[j][k] + (r1 * math.cos(r2) * abs(r3 * self.gbest_x[k] - self.X[j][k]))
                    except:
                        self.X[j][k] = self.X[j][k] + (r1 * math.cos(r2) * abs(r3 * self.gbest_x[0][k] - self.X[j][k]))
        self.X = np.clip(self.X, self.lb, self.ub)
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]  # Function for fitness evaluation of new solutions


    def run(self):
        for i in range(self.max_iter):
            self.update(i)
            self.update_pbest()
            self.update_gbest()
            self.gbest_y_hist.append(self.gbest_y)
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y

if __name__ == '__main__':
    n_dim = 2
    lb = [-512 for i in range(n_dim)]
    ub = [512 for i in range(n_dim)]
    demo_func = test_function.f23
    sca = sca(n_dim=2, pop_size=40, max_iter=150, lb=lb, ub=ub, func=demo_func)
    sca.run()
    print('best_x is ', sca.gbest_x, 'best_y is', sca.gbest_y)
    print(f'{demo_func(sca.gbest_x)}\t{sca.gbest_x}')
    plt.plot(sca.gbest_y_hist)
    plt.show()



# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/3 17:36
# @Author : LucXiong
# @Project : Model
# @File : SBO.py

"""
Ref:https://blog.csdn.net/u011835903/article/details/107857884
"""

import random  # random Function
import numpy as np # numpy operations
import matplotlib.pyplot as plt
import math
import test_function

class SBO():
    def __init__(self, pop_size=50, n_dim=2, alpha=0.94, lb=-1e5, ub=1e5, max_iter=1000, func=None):
        self.pop = pop_size
        self.n_dim = n_dim
        self.alpha = alpha # 步长的最大阈值
        self.func = func
        self.max_iter = max_iter  # max iter
        self.z = 0.02 # z是缩放比例因子
        self.r_mutate = 0.05  # 变异概率

        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        assert self.n_dim == len(self.lb) == len(self.ub), 'dim == len(lb) == len(ub) is not True'
        assert np.all(self.ub > self.lb), 'upper-bound must be greater than lower-bound'

        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.pop, self.n_dim))
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))] # y = f(x) for all particles

        self.pbest_x = self.X.copy()  # personal best location of every particle in history
        self.pbest_y = [np.inf for i in range(self.pop)]  # best image of every particle in history
        self.fit = [1 / (1 + self.Y[i]) if self.Y[i] > 0 else 1 - self.Y[i] for i in range(self.pop)]
        self.prob = [self.fit[i] / sum(self.fit) for i in range(self.pop)]
        self.gbest_x = self.pbest_x.mean(axis=0).reshape(1, -1)  # global best location for all particles
        self.gbest_y = np.inf  # global best y for all particles
        self.gbest_y_hist = []  # gbest_y of every iteration
        self.update_gbest()

    def update_pbest(self):
        '''
        personal best
        :return:
        '''
        for i in range(len(self.Y)):
            if self.pbest_y[i] > self.Y[i]:
                self.pbest_x[i] = self.X[i]
                self.pbest_y[i] = self.Y[i]

    def update_gbest(self):
        '''
        global best
        :return:
        '''
        idx_min = self.pbest_y.index(min(self.pbest_y))
        if self.gbest_y > self.pbest_y[idx_min]:
            self.gbest_x = self.X[idx_min, :].copy()
            self.gbest_y = self.pbest_y[idx_min]

    def cal_prob(self):
        self.fit = [1 / (1 + self.Y[i]) if self.Y[i] > 0 else 1 - self.Y[i] for i in range(self.pop)]
        self.prob = [self.fit[i] / sum(self.fit) for i in range(self.pop)]

    def update(self, iter_num):
        idx_min = self.Y.index(min(self.Y))
        for i in range(self.pop):
            # roulette wheel
            for k in range(self.n_dim):
                select_list = []
                while len(select_list) < 1:
                    select_list = []
                    r = np.random.rand(1)
                    for kk in range(len(self.prob)):
                        if self.prob[kk] > (r[0]):
                            select_list.append(kk)
                j = random.choice(select_list)
                lemta = self.alpha / (1 + self.prob[j])
                try:
                    self.X[i, k] += lemta * 0.5 * (self.X[j, k] + self.gbest_x[0][k] - 2 * self.X[i, k])
                except:
                    self.X[i, k] += lemta * 0.5 * (self.X[j, k] + self.gbest_x[k] - 2 * self.X[i, k])

            # 变异
            if np.random.rand(1) < self.r_mutate and i != idx_min:
                for j in range(self.n_dim):
                    # 正态分布 Satin bowerbird optimizer: A new optimization algorithm to optimize
                    # ANFIS for software development effort estimation.pdf
                    self.X[i][j] += self.z * np.random.normal() * (self.ub[j] - self.lb[j])
        self.X = np.clip(self.X, self.lb, self.ub)
        self.Y = [self.func(self.X[i]) for i in range(len(self.X))]

    # def mutate(self):
    #     # 没有用上这个函数
    #     idx_min = self.Y.index(min(self.Y))
    #     for i in range(self.pop):
    #         if np.random.rand(1)[0] < self.r_mutate and i != idx_min:
    #             for j in range(self.n_dim):
    #                 self.X[i][j] += self.z * np.random.normal() * (self.ub[j] - self.lb[j])
    #     self.X = np.clip(self.X, self.lb, self.ub)
    #     self.Y = [self.func(self.X[i]) for i in range(len(self.X))]

    def run(self):
        for i in range(self.max_iter):
            print(i)
            self.update(i)
            self.cal_prob()
            self.update_pbest()
            self.update_gbest()
            self.gbest_y_hist.append(self.gbest_y)
        self.best_x, self.best_y = self.gbest_x, self.gbest_y
        return self.best_x, self.best_y

if __name__ == '__main__':
    # 寻优效果不错，但是计算时间较长，有一部分原因是我频繁整体计算适应度值
    n_dim = 30
    lb = [-100 for i in range(n_dim)]
    ub = [100 for i in range(n_dim)]
    demo_func = test_function.fu5
    pop_size = 100
    max_iter = 100
    sbo = SBO(n_dim=n_dim, pop_size=pop_size, max_iter=max_iter, lb=lb, ub=ub, func=demo_func)
    best_x, bext_y = sbo.run()
    print(f'{demo_func(sbo.gbest_x)}\t{sbo.gbest_x}')
    plt.plot(sbo.gbest_y_hist)
    plt.show()




# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/6 10:44
# @Author : LucXiong
# @Project : Model
# @File : test_func.py

"""
https://blog.csdn.net/miscclp/article/details/38102831
"""

import math

# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/6 10:44
# @Author : LucXiong
# @Project : Model
# @File : test_func.py

"""
https://blog.csdn.net/miscclp/article/details/38102831
"""

import math
import random
from scipy.stats import norm
# Unimodal test functions
def fu1(x):
    # Sphere function
    # min is 0 at [0 for i in range(len(x)]
    s1 = 0
    for i in range(len(x)):
        s1 += (x[i] ** 2)
    return s1

def fu2(x):
    '''
    min is 0 at [0 for i in range(len(x)]
    :param x: xi ∈ [-10, 10]
    :return:
    '''

    s1 = 0
    s2 = 1
    for i in range(len(x)):
        s1 += abs(x[i])
        s2 *= abs(x[i])
    result = s1 + s2
    return result

def fu3(x):
    # min is 0 at [0 for i in range(len(x)]
    s1 = 0
    for i in range(len(x)):
        s2 = 0
        for j in range(i):
            s2 += abs(x[i])
        s1 += s2 ** 2
    result = s1
    return result

def fu4(x):
    # min is 0 at [0 for i in range(len(x)]
    y = []
    for i in range(len(x)):
        y.append(abs(x[i]))
    return max(y)

def fu5(x):
    '''
    min is 0 at [-0.5 for i in range(len(x)]
    :param x:
    :return:
    '''
    s1 = 0
    for i in range(len(x)):
        s1 += (abs(x[i] + 0.5) ** 2)
    result = s1
    return result

def fu6(x):
    ''' Rosenbrock function
        min is 0 at [1 for i in range(len(x)]
        :param x:
        :return:
        '''
    s1 = 0
    for i in range(len(x) - 1):
        s1 += (100 * (x[i+1] - x[i] ** 2) ** 2 + (x[i] - 1) ** 2)
    result = s1
    return result

def fu7(x):
    s1 = 0
    for i in range(len(x)):
        s1 += (i * x[i] ** 4 + norm.rvs())
    result = s1
    return result

# Multimodal test functions
def fm1(x): # Eggholder Function
    '''
    min is 0 at [0 for i in range(len(x)]
    :param x: xi ∈ [-5.12, 5.12]
    :return:
    '''
    result = 0
    for i in range(len(x)):
        result += (x[i] ** 2 - 10 * math.cos(2 * math.pi * x[i]) + 10)
    return result

def fm2(x):
    '''
    GRIEWANK FUNCTION:http://www.sfu.ca/~ssurjano/griewank.html
    :param x:-600,600
    :return:
    '''
    s1 = 0
    s2 = 1
    for i in range(len(x)):
        s1 += x[i] ** 2
        s2 *= math.cos(x[i] / math.sqrt(i + 1))
    s1 = s1 / 4000
    result = s1 - s2 + 1
    return result

def fm3(x):
    # min is −418.9829 * len(x)
    # x: -500, 500
    s1 = 0
    for i in range(len(x)):
        s1 += -x[i] * math.sin(math.sqrt(abs(x[i])))
    result = s1
    return result

def fm4(x):
    '''
    ACKLEY FUNCTION:http://www.sfu.ca/~ssurjano/ackley.html
    :param x:xi ∈ [-32.768, 32.768]
    :return:
    '''
    a = 20
    b = 0.2
    c = 2 * math.pi
    s1 = 0
    s2 = 0
    for i in range(len(x)):
        s1 += x[i] ** 2
        s2 += math.cos(c * x[i])
    s1 = -a * math.exp(-b * math.sqrt(s1 / len(x)))
    s2 = -math.exp(s2 / len(x))
    result = s1 + s2 + a + math.exp(1)
    return result

def fm5(x):
    # x: -50, 50
    def y(xi):
        y = 1 + (xi + 1) / 4
        return y

    def u(xi):
        a  = 10
        k = 100
        m = 4
        if xi > a:
            u = k * pow((xi -a), m)
        elif xi < -a:
            u = k * pow((-xi - a), m)
        else:
            u = 0
        return u

    s1 = 10 * math.sin(math.pi * y(x[0])) ** 2 + (y(x[-1]) - 1) ** 2
    # s1 = 10 * math.sin(math.pi * y(x[0])) + (y(x[-1]) - 1) ** 2
    # 在已有的算法文章中使用的关于fm5的测试函数s1 = 10 * math.sin(math.pi * y(x[0])) + (y(x[-1]) - 1) ** 2
    # Xue J, Shen B. A novel swarm intelligence optimization approach: sparrow search algorithm[J].
    # Systems Science & Control Engineering, 2020, 8(1): 22-34. 中的F12
    #Mirjalili S, Lewis A. The Whale Optimization Algorithm[J]. Advances in Engineering Software,
    # 2016, 95: 51-67. 中的F12
    # 但是在这种情况下，该测试函数的理论最小值显然不是0；因此怀疑文章关于此测试函数的描述出现了问题
    # 比如当x[0]=1,x[i]=-1(i!=0)时，此时的F12<0
    for i in range(len(x)-1):
        s1 += ((y(x[i]) - 1) ** 2 * (1 + 10 * math.sin(math.pi * y(x[i+1])) ** 2))
    s2 = 0
    for i in range(len(x)):
        s2 += u(x[i])
    result = s1 * math.pi / len(x) + s2
    return result

# . Fixed-dimension test functions
# two dimension
def f21(x):
    '''
    BUKIN FUNCTION N. 6:http://www.sfu.ca/~ssurjano/bukin6.html
    :param x: x1 ∈ [-15, -5], x2 ∈ [-3, 3].
    :return:
    '''
    s1 = 100 * math.sqrt(abs(x[1] - 0.01 * x[0] ** 2))
    s2 = 0.01 * abs(x[0] + 10)
    result = s1 + s2
    return result

def f22(x):
    # min is -1.0316
    x1 = x[0]
    x2 = x[1]
    result = 4 * x1 ** 2 - 2.1 * x1 ** 4 + x1 ** 6 /3 + x1 * x2 - 4 * x2 ** 2 + 4 * x2 ** 4
    return result

def f23(x):
    # Eggholder function:
    # xi ∈ [-512, 512]
    # min is f(512, 404.2319) =-959.6407
    x1 = x[0]
    x2 = x[1]
    result = -(x2 + 47) * math.sin(math.sqrt(math.fabs(x1 + x2/2 + 47))) - x1 * math.sin(math.sqrt(math.fabs(x1 - x2 - 47)))
    return result

def f24(x):
    # Ackley's function
    # xi ∈ [-5, 5]
    # min is f(0, 0) = 0
    x1 = x[0]
    x2 = x[1]
    s1 = -20 * math.exp(-math.sqrt((x1 ** 2 + x2 ** 2)/50))
    s2 = math.exp(0.5 * math.cos(2 * math.pi * x1) + math.cos(2 * math.pi * x2))
    result = s1 - s2 + 20 + math.e
    return result

def f25(x):
    # Beale 's function
    # xi ∈ [-4.5, 4.5]
    # min is f(3, 0.5) = 0
    x1 = x[0]
    x2 = x[1]
    s1 = (1.5 - x1 + x1 * x2) ** 2
    s2 = (2.25 - x1 + x1 * x2 * x2) ** 2
    s3 = (2.625 - x1 + x1 * x2 * x2 * x2) ** 2
    result = s1 + s2 + s3
    return result

def f26(x):
    # Lévi function N.13
    # xi ∈ [-10, 10]
    # min is f(1, 1) = 0
    x1 = x[0]
    x2 = x[1]
    s1 = (math.sin(3 * math.pi * x1)) ** 2
    s2 = (x1 - 1) * (x1 - 1) * (1 + (math.sin(3 * math.pi * x2)) ** 2)
    s3 = (x2 - 1) * (x2 - 1) * (1 + (math.sin(3 * math.pi * x2)) ** 2)
    result = s1 + s2 + s3
    return result

def f27(x):
    # Cross-in-tray function:
    # xi ∈ [-10, 10]
    # min is f(±1.34941, ±1.34941) = -2.06261
    x1 = x[0]
    x2 = x[1]
    s1 = abs(100 - math.sqrt(x1 * x1 + x2 * x2) / math.pi)
    s2 = math.sin(x1) * math.sin(x2) * math.exp(s1)
    s3 = -0.0001 * math.pow((abs(s2) + 1), 0.1)
    result = s3
    return result

if __name__ == '__main__':
    pass


# ！usr/bin/env python
# -*- coding: utf-8 -*-
# Time : 2021/12/15 19:52
# @Author : LucXiong
# @Project : Model
# @File : setup.py
# https://blog.csdn.net/weixin_49246443/article/details/108425035
from setuptools import setup, find_packages

setup(
    name="swarm_algorithm",
    version="0.0.2",
    author="Luc",
    author_email="xionglei@sjtu.edu.cn",
    description="",
    long_description_content_type="text/markdown",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
    install_requires=[],  # install_requires字段可以列出依赖的包信息，用户使用pip或easy_install安装时会自动下载依赖的包
    url='https://github.com',
    license='MIT',
    packages=find_packages(),  # 需要处理哪里packages，当然也可以手动填，例如['pip_setup', 'pip_setup.ext']
    include_package_data=False,
    zip_safe=True,
)